#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef enum {BIG_ENDIAN, LITTLE_ENDIAN} tOrdenacao;

/****
 *
 * Funo OrdenacaoDeBytes()
 *
 * Descrio: Verifica se o sistema corrente 
 *            big ou little-endian
 *
 * Parmetros: Nenhum
 *
 * Retorno: Valor indicando se o sistema corrente 
 *          big ou little-endian.
 *
 ****/

tOrdenacao OrdenacaoDeBytes(void)
{
   int  x = 1;
   char primeiroByte;

   primeiroByte = *(char *)&x == 1;

   return primeiroByte ? LITTLE_ENDIAN : BIG_ENDIAN;
}

/****
 *
 * Funo EscreveBigEndian()
 *
 * Descrio: Esta funo  semelhante a fwrite(), mas
 *            escreve sempre em big-endian.
 *
 * Parmetros:
 *      array (entrada): ponteiro para um array no qual
 *                       os bytes sero lidos
 *      tamanhoElemento (entrada): tamanho de cada
 *                                 elemento do array
 *      nElementos (entrada): nmero de itens do array
 *                            a ser escritos
 *      stream(entrada): onde ser feita a escrita
 *
 * Retorno: nmero de itens (no  o nmero de bytes!)
 *          que foram realmente escritos no stream.
 *
 ****/

size_t EscreveBigEndian(const void *array,
                        size_t tamanhoElemento,
                        size_t nElementos, FILE *stream)
{
   size_t i, j, retorno = 0;
   char   *copia, *buffer;

      /* Se a ordenao j  big-endian, basta chamar */
      /* a funo fwrite() para realizar a tarefa.    */
      /* Se a ordenao de bytes desejada for little- */
      /* endian,  esta  a nica instruo que        */
      /* precisa ser alterada.                        */
   if (OrdenacaoDeBytes() == BIG_ENDIAN)
      return fwrite( array, tamanhoElemento, nElementos,
                     stream );

         /* O bloco apontado por copia conter       */
         /* uma cpia de cada elemento do array.     */
         /* Este bloco no  essencial, mas facilita */
   copia = calloc(1, tamanhoElemento);

         /* O bloco apontado por buffer     */
         /* conter uma cpia de cada       */
         /* elemento do array em big-endian */
   buffer = calloc(1, tamanhoElemento);

      /* Se no for possvel alocar os dois   */
      /* blocos, desiste, mas essa estratgia */
      /* pode ser melhorada.                  */
   if (!copia || !buffer)
      return 0;

         /* Escreve um a um cada elemento do */
         /* array com os bytes ordenados em  */
         /* formato big-endian               */
   for (i = 0; i < nElementos; ++i) {
         /* Primeiro faz um cpia de cada elemento. */
         /* Este passo no  essencial, mas ajuda.  */
      memmove(copia, (char *)array + i*tamanhoElemento,
               tamanhoElemento);

         /* Armazena cada byte de cada elemento */
         /* do array original em ordenao big- */
         /* endian no array buffer.             */
      for (j = 0; j < tamanhoElemento; ++j)
         buffer[tamanhoElemento - j - 1] = copia[j];

         /* A funo fwrite() escreve os bytes no     */
         /* arquivo na ordem em que eles se encontram */
         /* em memria. Como o bloco buffer contm os */
         /* bytes do elemento corrente do array em    */
         /* ordenao big-endian assim sero escrito  */
         /* os bytes no arquivo.                      */
      retorno += fwrite( buffer, tamanhoElemento, 1,
                         stream );

         /* Se ocorrer algum erro de escrita retorna */
         /* o nmero de bytes escritos at ento.    */
      if (ferror(stream))
         goto tchau; /* Um goto desses no faz mal */
   }

tchau:
   free(copia);
   free(buffer);

     /* Se no ocorreu nenhum erro, o valor retornado */
     /* neste ponto dever ser igual a nElementos     */
   return retorno;
}

/****
 *
 * Funo LeBigEndian()
 *
 * Descrio: Esta funo  semelhante a fread(), mas
 *            l sempre em big-endian.
 *
 * Parmetros:
 *      array (entrada): ponteiro para um array no qual
 *                      os bytes lidos sero depositados
 *      tamanhoElemento (entrada): tamanho de cada
 *                                 elemento lido
 *      nElementos (entrada): nmero de itens do array
 *                            que sero lidos
 *      stream(entrada): onde a leitura ser feita
 *
 * Retorno: nmero de itens (no  o nmero de bytes!)
 *          que foram realmente lidos no stream.
 *
 * OBSERVAO:  assumido que a ordenao do arquivo
 *              big-endian. No h como esta funo
 *             verificar se isto  verdadeiro ou no
 *
 ****/

size_t LeBigEndian( void *array, size_t tamanhoElemento,
                    size_t nElementos, FILE *stream)
{
   size_t i, j, retorno = 0;
   char   *copia, *buffer;

      /* Se a ordenao j  big-endian, basta chamar */
      /* a funo fread() para realizar a tarefa.     */
      /* Se a ordenao de bytes desejada for little  */
      /* -endian, esta  a nica instruo que        */
      /* precisa ser alterada.                        */
   if (OrdenacaoDeBytes() == BIG_ENDIAN)
      return fread( array, tamanhoElemento, nElementos,
                    stream );

         /* O bloco apontado por copia conter       */
         /* uma cpia de cada elemento do array.     */
         /* Este bloco no  essencial, mas facilita */
   copia = calloc(1, tamanhoElemento);

         /* O bloco apontado por buffer        */
         /* conter uma cpia de cada elemento */
         /* do array em big-endian             */
   buffer = calloc(1, tamanhoElemento);

      /* Se no for possvel alocar os  */
      /* dois blocos, desiste, mas isso */
      /* pode ser melhorado.            */
   if (!copia || !buffer)
      return 0;


         /* L um a um cada elemento do array e   */
         /* ordena os bytes em formato big-endian */
   for (i = 0; i < nElementos; ++i) {
         /* L cada elemento no bloco copia. */
      retorno += fread( copia, tamanhoElemento, 1,
                        stream );

         /* Se ocorrer algum erro de leitura retorna */
         /* o nmero de bytes escritos at ento     */
      if (ferror(stream)) {
         free(copia);
         free(buffer);
         return retorno;
      }

         /* Armazena cada byte de cada elemento */
         /* do array original em ordenao big- */
         /* endian no array buffer.             */
      for (j = 0; j < tamanhoElemento; ++j)
         buffer[tamanhoElemento - j - 1] = copia[j];

      memmove( (char *)array + i*tamanhoElemento,
               buffer, tamanhoElemento );
   }

   free(copia);
   free(buffer);

      /* Se no ocorreu nenhum erro, o */
      /* valor retornado neste ponto   */
      /* dever ser igual a nElementos */
   return retorno;
}

int main(void)
{
   double x = 2.54, y = 0, z = 0;
   FILE   *stream = fopen("Arq.dat", "wb");

   if (!stream) {
      printf("Erro ao abrir arquivo para escrita\n");
      return 1;
   }

   if (1 != EscreveBigEndian(&x, sizeof(x), 1, stream)){
      printf("Ocorreu um erro de escrita no arquivo\n");
      return 1;
   }

   fclose(stream);

   stream = fopen("Arq.dat", "rb");

   if (!stream) {
      printf("Erro ao abrir arquivo para leitura\n");
      return 1;
   }

   if (1 != fread(&y, sizeof(y), 1, stream)) {
      printf("Ocorreu um erro de leitura no arquivo\n");
      return 1;
   }

      /*  preciso retorna ao incio do arquivo */
   rewind(stream);

   if (1 != LeBigEndian(&z, sizeof(z), 1, stream)) {
      printf("Ocorreu um erro de leitura no arquivo\n");
      return 1;
   }

   printf("Valor original:               %4.2lE\n", x);
   printf("Valor lido com fread():       %4.2lE\n", y);
   printf("Valor lido com LeBigEndian(): %4.2lE\n", z);

   return 0;
}

/*****************************************

Resultado do program:

Valor original:               2.54E+000
Valor lido com fread():       3.07E+090
Valor lido com LeBigEndian(): 2.54E+000

*****************************************/
